/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File xsamgrx.c : SAM display driver routines - X11 version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#if defined(nec_ews) && !defined(__STDC__)
void *shmat(int, void*, int);	/* shut the compiler up */
#endif
#endif


#include "z80.h"
#include "mem.h"
#include "sam.h"
#include "util.h"
#include "xsamP.h"

static uns8 *x_data_ptr;
static XImage *xim;
#ifdef MITSHM
static XShmSegmentInfo xshminfo;
#endif

extern Display *dpy;
extern int scrn;
extern Window root;
extern Window mainWin;
extern GC theGc;

extern int flash_phase;
extern int ClutChangeInFrame;
extern int BorderChangeInFrame;
extern int VmprChangeInFrame;	

/* These macros map a byte in the attribute area to an vga colour index */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)

uns8 LineNeedsUpdating[256];	/* Determines whether a given line needs to
				   be redrawn */

static uns8 *SamMemoryPtr;	/* Pointer to the SAM screen memory */
static uns8 *XMemoryPtr;	/* Pointer within the X frame buffer */

void (*borderline_update)(void);
void (*line_update)(int);
void (*line_update_mode_1)(int);
void (*line_update_mode_2)(int);
void (*line_update_mode_3)(int);
void (*line_update_mode_4)(int);

void (*screen_refresh)(void);
void (*screen_refresh_1x1)(void);
void (*screen_refresh_2x2)(void);

int BorderTop;
int BorderBottom;

void
borderline_update_1x1(void)
{
/* Update either top or bottom borders if required */

   if (ClutChangeInFrame || BorderChangeInFrame)
   	memset(XMemoryPtr, clut[BORD_VAL(border)], HSize1x1);
   
   XMemoryPtr += HSize2x2;
   	
}

void
line_update_x_mode1_1x1(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 1 line */
	XMemoryPtr += HSize2x2;	
   }
}

void
line_update_x_mode2_1x1(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr += HSize2x2;
   }
}


void
line_update_x_mode3_1x1(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */
	XMemoryPtr += HSize2x2;
   }
}

void
line_update_x_mode4_1x1(int LineNo)
{
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_1x1);
	XMemoryPtr += BorderWidth_H_1x1;

	/* Skip rest of XImage line, as we only fill half for 1x1 */
	XMemoryPtr += HSize1x1;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */
	XMemoryPtr += HSize2x2;
   }
}


void
borderline_update_2x2(void)
{
/* Update either top or bottom borders if required */

   if (ClutChangeInFrame || BorderChangeInFrame)
   	memset(XMemoryPtr, clut[BORD_VAL(border)], 2*HSize2x2);
   
   XMemoryPtr += 2*HSize2x2;
   	
}


void
line_update_x_mode1_2x2(int LineNo)
{

   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8 *XMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

   extern unsigned int y_dtbl[];

/* In Mode 1 we have to update SamMemoryPtr each line due to the 
   spectrum screen layout, but we use a lookup table to minimize
   the cost */

   SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
   SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
   
/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr;

	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++;
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}

		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy(XMemoryPtr, XMemoryPtrSave, HSize2x2);
	XMemoryPtr += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}
void
line_update_x_mode2_2x2(int LineNo)
{
   unsigned int ink, paper, tmp;
   uns8 attr, i_pix, p_pix;
   uns8 *XMemoryPtrSave;
   uns8* SamAttrPtr;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr once in a given frame, at LineNo == 0 */
   
   if (VmprChangeInFrame || !LineNo)
   	SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   


  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr;
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	SamAttrPtr = SamMemoryPtr + 0x2000;
	for (i=0; i < 32; i++) {
		attr = *SamAttrPtr++; 
		ink = Foreground(attr); paper = Background(attr);

		/* toggle the colours if we're in the inverse part of the FLASH cycle */
		if (flash_phase && (attr & 0x80)) {
			tmp=ink; ink=paper; paper=tmp;
		}
		i_pix = clut[ink];
		p_pix = clut[paper];
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x80) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x40) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x20) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x10) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x08) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x04) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x02) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*SamMemoryPtr & 0x01) ? i_pix : p_pix;
		*(XMemoryPtr++) = (*(SamMemoryPtr++) & 0x01) ? i_pix : p_pix;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy(XMemoryPtr, XMemoryPtrSave, HSize2x2);
	XMemoryPtr += HSize2x2;

	LineNeedsUpdating[LineNo] = False;


   }
   else {
   /* Short route */
	SamMemoryPtr += 32;		/* 32 Bytes in a mode 2 line */
	XMemoryPtr += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}

void
line_update_x_mode3_2x2(int LineNo)
{
   uns8 *XMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }

/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
       BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   
   
  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr;

	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0xc0) >> 6)];
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x30) >> 4)];
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x0c) >> 2)];
		*(XMemoryPtr++) = clut[HMPR_MD3COL | MD3_SWITCH((*SamMemoryPtr & 0x03))];
		SamMemoryPtr++;
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* Now generate the 2nd line by copying the first */
	memcpy(XMemoryPtr, XMemoryPtrSave, HSize2x2);
	XMemoryPtr += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 3 line */
	XMemoryPtr += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}

void
line_update_x_mode4_2x2(int LineNo)
{
   uns8 *XMemoryPtrSave;
   int i;

/* If there is no flagged vmpr change then we only need to update 
   SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
   at LineNo == 128 (when we cross the page boundry ) */
   
   if (VmprChangeInFrame || !LineNo || (LineNo == 128)) {
   	SamMemoryPtr = (LineNo & 0x80) ? 
   			realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
   			realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
   			   
   }



/* Now we have a long or a short route either we need to update the frame
   buffer (if the memory has been written to, or we have a vmpr/clut change
   to deal with), or we just need to update the read/write pointers */
   
   if (VmprChangeInFrame || ClutChangeInFrame || 
     BorderChangeInFrame || LineNeedsUpdating[LineNo]) {
   /* Long route */   

  	/* Save the X pointer - we need this for generating the 
  	   next line down - using memcpy */

	XMemoryPtrSave = XMemoryPtr;
   
	/* Lhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;

	/* The Actual screen line */
	
	for (i=0; i < 128; i++) {
		*(XMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr++) = clut[((*SamMemoryPtr)   &0xf0) >> 4];
		*(XMemoryPtr++) = clut[ (*SamMemoryPtr)   &0x0f];
		*(XMemoryPtr++) = clut[ (*SamMemoryPtr++) &0x0f];
	}
	
	/* Rhs Border */
	memset(XMemoryPtr, clut[BORD_VAL(border)], BorderWidth_H_2x2);
	XMemoryPtr += BorderWidth_H_2x2;


	/* Now generate the 2nd line by copying the first */
	memcpy(XMemoryPtr, XMemoryPtrSave, HSize2x2);
	XMemoryPtr += HSize2x2;

	LineNeedsUpdating[LineNo] = False;

   }
   else {
   /* Short route */
	SamMemoryPtr += 128;		/* 128 Bytes in a mode 4 line */
	XMemoryPtr += 2*HSize2x2;	/* No. Bytes in a x 2x2 line */
   }
}

#ifdef MITSHM
/* Screen refresh functions - two different functions for each
   scale mode, and one set for either shared or non shared transport */
void 
screen_refresh_xshm_1x1(void)
{

	XShmPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize1x1, VSize1x1, False);
	XFlush(dpy);
}
void 
screen_refresh_xshm_2x2(void)
{

	XShmPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize2x2, VSize2x2, False);
	XFlush(dpy);
}
#endif

void 
screen_refresh_xim_1x1(void)
{

	XPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize1x1, VSize1x1);
	XFlush(dpy);
}
void 
screen_refresh_xim_2x2(void)
{

	XPutImage(dpy, mainWin, theGc, xim, 0, 0, 0, 0,
		HSize2x2, VSize2x2);
	XFlush(dpy);
}


void
FlyBack(void)
{
/* Reset the X frame buffer pointer, and refresh screen */

   XMemoryPtr = x_data_ptr;
   screen_refresh();

}


void
ChangeScaleMode(int SF)
{

	XSizeHints xsh;
	int i;
	
	BorderTop    = 5; 
	BorderBottom = 246;
	
	if (SF == 1) {
	
		borderline_update      = borderline_update_1x1;
		line_update_mode_1 = line_update_x_mode1_1x1;
		line_update_mode_2 = line_update_x_mode2_1x1;
		line_update_mode_3 = line_update_x_mode3_1x1;
		line_update_mode_4 = line_update_x_mode4_1x1;

		xsh.min_width  = HSize1x1;
		xsh.min_height = VSize1x1;
		xsh.max_width  = HSize1x1;
		xsh.max_height = VSize1x1;


		screen_refresh       = screen_refresh_1x1;
		XResizeWindow(dpy, mainWin, HSize1x1, VSize1x1);

	}
	else {
	
		borderline_update      = borderline_update_2x2;
		line_update_mode_1 = line_update_x_mode1_2x2;
		line_update_mode_2 = line_update_x_mode2_2x2;
		line_update_mode_3 = line_update_x_mode3_2x2;
		line_update_mode_4 = line_update_x_mode4_2x2;

		xsh.min_width  = HSize2x2; 
		xsh.min_height = VSize2x2;
		xsh.max_width  = HSize2x2;
		xsh.max_height = VSize2x2;

		screen_refresh       = screen_refresh_2x2;
		XResizeWindow(dpy, mainWin, HSize2x2, VSize2x2);

	}

	/* Notify the Window Manager of the change of size */

	xsh.flags = PSize | PMinSize | PMaxSize;
	XSetWMNormalHints(dpy, mainWin, &xsh);	

	/* Make sure we update the whole screen */
	for (i=0; i < 192; i++)
	LineNeedsUpdating[i] = True;
	BorderChangeInFrame = (192+59);
	FlyBack();
}

/*
 * Image routines
 */


int
open_image(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		xim = XShmCreateImage(dpy, DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, NULL, &xshminfo, HSize2x2, VSize2x2+8);
		if (!xim) {
			simcoupe_mesg(SIMCOUPE_ERR, "SimCoupe: failed to create a shared XImage!");
			return -1;
		}
		xshminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line * xim->height,
			IPC_CREAT|0777);
		if (xshminfo.shmid == -1) {
			simcoupe_mesg(SIMCOUPE_ERR, "shmget <xshminfo.shmid> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.shmaddr = xim->data = shmat(xshminfo.shmid, 0, 0);
		if (!xshminfo.shmaddr) {
			simcoupe_mesg(SIMCOUPE_ERR, "shmat <xshminfo.shmaddr> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.readOnly = False;
		if (!XShmAttach(dpy, &xshminfo)) {
			simcoupe_mesg(SIMCOUPE_ERR, "XShmAttach failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh_1x1 = screen_refresh_xshm_1x1;
		screen_refresh_2x2 = screen_refresh_xshm_2x2;
	} else
#endif
	{
		xim = XCreateImage(dpy,  DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, 0, NULL, HSize2x2, VSize2x2+8, 8, 0);
		if (!xim) {
			simcoupe_mesg(SIMCOUPE_ERR, "XCreateImage failed");
			perror("  reason:");
			return -1;
		}
		if ((xim->data = (char *)malloc(xim->bytes_per_line * xim->height))
			== NULL) {
			simcoupe_mesg(SIMCOUPE_ERR, "malloc <xim->data> failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh_1x1 = screen_refresh_xim_1x1;
		screen_refresh_2x2 = screen_refresh_xim_2x2;
	}
	x_data_ptr = xim->data;

	return 0;
}

void
close_image(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		XShmDetach(dpy, &xshminfo);
		XDestroyImage(xim);
		shmdt(xshminfo.shmaddr);
		shmctl(xshminfo.shmid, IPC_RMID, 0);
	}
	else
#endif
		XDestroyImage(xim);
}

